@@ -10,6 +10,7 @@ use RPC::ExtDirect ();       # No imports needed here
 use RPC::ExtDirect::Serialize;
 use RPC::ExtDirect::Event;
 use RPC::ExtDirect::NoEvents;
+use RPC::ExtDirect::Hook;
 
 ### PACKAGE GLOBAL VARIABLE ###
 #
@@ -26,7 +27,7 @@ our $DEBUG = 0;
 #
 
 sub poll {
-    my ($class) = @_;
+    my ($class, $env) = @_;
 
     # First set the debug flag
     local $RPC::ExtDirect::Serialize::DEBUG = $DEBUG;
@@ -35,35 +36,86 @@ sub poll {
     my @handler_refs = RPC::ExtDirect->get_poll_handlers();
 
     # Even if we have nothing to poll, we must return a stub Event
+    # or client side will throw an unhandled JavaScript exception
     return $class->_no_events unless @handler_refs;
 
     # Compile the list of poll handler references
-    my @code_refs;
+    my @poll_handlers;
     for my $handler_ref ( @handler_refs ) {
-        my $action = $handler_ref->[0];
-        my $method = $handler_ref->[1];
+        my ($action, $method) = @$handler_ref;
 
         my %params
             = RPC::ExtDirect->get_method_parameters($action, $method);
 
-        push @code_refs, $params{referent};
+        push @poll_handlers, \%params;
     };
 
     # Run all the handlers and collect their outputs
     my @results;
-    CODE_REF:
-    for my $code_ref ( @code_refs ) {
-        next CODE_REF unless defined $code_ref;
 
-        my @output = eval { $code_ref->() };
+    POLL_HANDLER:
+    for my $handler ( @poll_handlers ) {
+        next POLL_HANDLER unless defined $handler;
 
-        # XXX Presently there is no way to return an exception to the
+        my (@output, $exception, $method_called);
+        my $run_method = 1;
+
+        # Run "before" hook if we got one
+        my $before = RPC::ExtDirect::Hook->new('before', $handler);
+
+        if ( $before ) {
+            my @hook_output = eval { $before->run($env, [$env]) };
+
+            # If "before" hook died, cancel Method call
+            if ( $@ ) {
+                $exception  = $@;
+                $run_method = '';
+            };
+
+            # If "before" hook returns anything but single number 1,
+            # or if it dies, we treat this as the result and do not
+            # run the poll handler.
+            if ( @hook_output != 1 || $hook_output[0] ne '1' ) {
+                @output = @hook_output;
+                $run_method = '';
+            };
+        };
+
+        # If "instead" hook is defined, we run it in place of handler
+        my $instead = RPC::ExtDirect::Hook->new('instead', $handler);
+
+        my $package  = $handler->{package};
+        my $referent = $handler->{referent};
+
+        if ( $run_method ) {
+            $method_called = $instead ? $instead->instead
+                           :            $referent
+                           ;
+
+            @output = $instead ? eval { $instead->run($env, [$env]) }
+                    :            eval { $referent->($package, $env) }
+                    ;
+
+            $exception = $@;
+        };
+
+        # Finally, run the "after" hook if it's defined
+        my $after = RPC::ExtDirect::Hook->new('after', $handler);
+
+        if ( $after ) {
+
+            # Return value and exceptions are ignored
+            eval {
+                $after->run($env, [$env], \@output, $exception, $method_called)
+            };
+        };
+
+        # Presently there is no way to return an exception to the
         # client side: Ext.Direct PollingProvider code does not have any
         # processing for exceptions and just freaks out upon receiving
         # anything but an event. Passing exceptions disguised as events
         # is kludgy and kinda defeats the whole purpose IMHO.
         # So in fact we just ignore anything exceptional on our side.
-        # *SIGH*
         push @results, eval { map { $_->result() } @output }
             unless $@;
     };
@@ -105,6 +157,27 @@ sub _no_events {
     return $serialized;
 }
 
+### PRIVATE PACKAGE SUBROUTINE ###
+#
+# Run specified hook
+#
+
+sub _run_hook {
+    my ($hook, $handler, $env, $output, $exception) = @_;
+
+    my %params = %$handler;
+
+    # Poll handlers are only passed env object as parameter
+    $params{arg} = [ $env ];
+
+    $params{code} = delete $params{referent};
+    $params{orig} = sub {
+        my ($code, $package, $arg) = @params{ qw/code package arg/ };
+
+        return $code->($package, $arg);
+    };
+}
+
 1;
 
 __END__
@@ -125,9 +198,10 @@ Alexander Tokarev E<lt>tokarev@cpan.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (c) 2011 Alexander Tokarev.
+Copyright (c) 2011-2012 Alexander Tokarev.
 
 This module is free software; you can redistribute it and/or modify it under
 the same terms as Perl itself. See L<perlartistic>.
 
 =cut
+
