@@ -1,19 +1,25 @@
 package RPC::ExtDirect::Test::Foo;
 
+use strict;
+use warnings;
+no  warnings 'uninitialized';
+
 use RPC::ExtDirect;
 
 # Return scalar result
-sub foo_foo : ExtDirect(1) {
+sub foo_foo : ExtDirect(1, before => \&foo_before) {
     return "foo! '${_[1]}'"
 }
 
 # Return arrayref result
-sub foo_bar : ExtDirect(2) {
-    return [ 'foo! bar!', $_[1], $_[2] ]
+sub foo_bar
+    : ExtDirect(2, instead => \&foo_instead)
+{
+    return [ 'foo! bar!', @_[1, 2], ];
 }
 
 # Return hashref result
-sub foo_baz : ExtDirect( params => [foo, bar, baz] ) {
+sub foo_baz : ExtDirect( params  => [foo, bar, baz], before  => \&foo_before, after   => \&foo_after) {
     my $class = shift;
     my %param = @_;
 
@@ -21,7 +27,7 @@ sub foo_baz : ExtDirect( params => [foo, bar, baz] ) {
                 bar => $param{bar},      baz => $param{baz},
               };
 
-    delete @param{ qw(foo bar baz) };
+    delete @param{ qw(foo bar baz _env) };
     @$ret{ keys %param } = values %param;
 
     return $ret;
@@ -41,4 +47,18 @@ sub foo_blessed : ExtDirect(0) {
     return bless {}, 'foo';
 }
 
+# Testing hooks
+sub foo_before {
+    return 1;
+}
+
+sub foo_instead {
+    my ($class, %params) = @_;
+
+    return $params{orig}->();
+}
+
+sub foo_after {
+}
+
 1;
